TITLE 80386 eForth for Linux 2.0.x

PAGE 62,132	;62 lines per page, 132 characters per line

.386p

;===============================================================
;
;	eForth 1.0 by Bill Muench and C. H. Ting, 1990
;	Much of the code is derived from the following sources:
;		8086 figForth by Thomas Newman, 1981 and Joe Smith, 1983
;		aFORTH by John Rible
;		bFORTH by Bill Muench
;
;	The goal of this implementation is to provide a simple eForth Model
;	which can be ported easily to many 8, 16, 24 and 32 bit CPU's.
;	The following attributes make it suitable for CPU's of the '90:
;
;		small machine dependent kernel and portable high level code
;		source code in the MASM format
;		direct threaded code
;		separated code and name dictionaries
;		simple vectored terminal and file interface to host computer
;		aligned with the proposed ANS Forth Standard
;		easy upgrade path to optimize for specific CPU
;
;	You are invited to implement this Model on your favorite CPU and
;	contribute it to the eForth Library for public use. You may use
;	a portable implementation to advertise more sophisticated and
;	optimized version for commercial purposes. However, you are
;	expected to implement the Model faithfully. The eForth Working
;	Group reserves the right to reject implementation which deviates
;	significantly from this Model.
;
;	As the ANS Forth Standard is still evolving, this Model will
;	change accordingly. Implementations must state clearly the
;	version number of the Model being tracked.
;
;	Representing the eForth Working Group in the Silicon Valley 
;	FIG Chapter. Send contributions to:
;
;		Dr. C. H. Ting
;		156 14th Avenue
;		San Mateo, CA 94402
;		(415) 571-7639
;
;
; This: '386 CPU or better, OS = Linux. 
; Marcel Hendrix, 26 October 1995
;
;===============================================================

;; Version control

VER		EQU	2			; major release version
EXT		EQU	0			; minor extension

;; Constants

TRUEE		EQU	-1			; true flag
FALSEE		EQU	 0			; false flag

COMPO		EQU	040H			; lexicon compile only bit
IMEDD		EQU	080H			; lexicon immediate bit
MASKK		EQU	07F7F7F1FH		; lexicon bit mask

CELLL		EQU	4			; size of a cell
BASEE		EQU	10			; default radix
VOCSS		EQU	8			; depth of vocabulary stack

BKSPP		EQU	127			; IN:  back space 
						; OUT: move back 1 char (\b)
CRR		EQU	10			; IN:  end of input
						; OUT: carriage return (\n)
PACECHAR	EQU	11
ERRO		EQU	128			; error message to host
CALLL		EQU	0E8909090h		; NOP CALL opcodes


_Readkbd	EQU	4			; Service offsets
_TTYOut		EQU	8
_ExitJob	EQU	12
_Type		EQU	16
_System		EQU	20


;;
;;


; These four can be freely changed

US		EQU	256*CELLL		; user area size in cells
RTS		EQU	1024*CELLL		; return stack/TIB size
DICTSIZE	EQU	0E0000h			; dictionary size ( 7/8 MByte)
NAMESIZE	EQU	02000h			; namelist size

;; Initialize assembly variables

_LINK	= 0					; force a null link
_USER	= 4*CELLL				; first user variable offset
_CODE	= CODEE					; initialize code pointer
_NAME	= DICTSIZE+16*CELLL+2*RTS+US+NAMESIZE	; initialize name pointer
_LEN	= 0

;; Define assembly macros

;	Compile a code definition header.

$CODE	MACRO	lex,name,label
	ALIGN	CELLL			;; force to cell boundary
label:					;; assembly label
	_CODE = $			;; save code pointer
	_LEN  = (lex AND 01FH)/CELLL	;; string cell count, round down
	_NAME = _NAME-((_LEN+3)*CELLL)	;; new header on cell boundary
ORG	_NAME+ORIG			;; set name pointer
	DD	 _CODE,_LINK		;; token pointer and link
	_LINK	= $			;; link points to a name string
	DB	lex,name		;; name string
ORG	_CODE				;; restore code pointer
	ENDM

;	Compile a headerless code definition.

$NOHEAD	MACRO	label
	ALIGN	CELLL		;; force to cell boundary
label:				;; assembly label
	ENDM

;	Compile a colon definition header.

$COLON	MACRO	lex,name,label
	$CODE	lex,name,label
	lea	ebx,4[ebx]	;; 3 byte NOP, spills ebx, 1 cycle
	call	DOLST		;; include CALL doLIST
	ENDM

;	Compile a headerless colon definition.

$NONAME	MACRO	label
	$NOHEAD	label
	lea	ebx,4[ebx]	;; 3 byte NOP, spills ebx!
	call	DOLST
	ENDM

;	Compile a user variable header.

$USER	MACRO	lex,name,label
	$CODE	lex,name,label
	lea	ebx,4[ebx]	;; 3 byte NOP, spills ebx!
	call	DOLST		;; include CALL doLIST
	DD	DOUSE,_USER	;; followed by doUSER and offset
	_USER = _USER+CELLL	;; update user area offset
	ENDM


;	Assemble inline direct threaded code ending.

$NEXT	MACRO
	lodsd			;; read the next code address into EAX
	jmp	eax		;; jump directly to the code address
	ENDM

; Call the Linux services

OSCALL	MACRO service				;; ( ? -- ? )
	mov	eax, dword ptr [JMPTABLE]
	call	service[eax]			;; call [service]
	ENDM


;; Main entry points and COLD start data


	IF MyOrg EQ 1000h	; MyOrg will be 1000h or 2000h

DFSEG 	SEGMENT
	DB 1000h DUP (0)
ENDS

	ENDIF

EFSEG	SEGMENT 
	ASSUME CS:EFSEG,DS:EFSEG,ES:EFSEG,SS:EFSEG 

						; beginning of cold boot area
		ORG MyOrg			; command line label

ORIG:		pop	eax			; return address
		pop	eax			; argc
		pop	eax			; argv
		pop	eax			; call table
		mov	dword ptr [JMPTABLE],eax
		mov	esp, OFFSET SPP		; initialize SP
		mov	ebp, OFFSET RPP		; initialize RP
		cld				; SI gets incremented
		mov	esi, OFFSET COLD1
		$NEXT				; to high level cold start



; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.

ALIGN CELLL					;align to cell boundary

UZERO:		DD	4 DUP (0)		; reserved space in user area
		DD	SPP			; SP0
		DD	RPP			; RP0
		DD	QRX			; '?KEY
		DD	TXSTO			; 'EMIT
		DD	ACCEP			; 'EXPECT
		DD	KTAP			; 'TAP
		DD	TXSTO			; 'ECHO
		DD	DOTOK			; 'PROMPT
		DD	BASEE			; BASE
		DD	0			; tmp
		DD	0			; SPAN
		DD	0			; >IN
		DD	0			; #TIB
		DD	TIBB			; TIB
		DD	0			; CSP
		DD	INTER			; 'EVAL
		DD	NUMBQ			; 'NUMBER
		DD	0			; HLD
		DD	0			; HANDLER
RETCOD:		DD	0			; RETURNCODE
		DD	0			; T0
		DD	0			; last
		DD	0			; penult
		DD	0			; STATE
		DD	doQBRAN			; (?BRANCH)
		DD	0			; CONTEXT pointer
		DD	VOCSS DUP (0)		; vocabulary stack
		DD	0			; CURRENT pointer
		DD	0			; vocabulary link pointer
		DD	CTOP			; CP
		DD	NTOP			; NP
		DD	LASTN			; LAST
ULAST:
		ORG ORIG+US			; beginning of code dictionary

JMPTABLE:	DD	0			; Linux jump table address

CODEE:

;; Device dependent I/O

;   TERMINATE	( n -- )
;		Exit eForth, return a value n.

		$CODE	9,'TERMINATE',TERMINATE
		OSCALL _ExitJob

;   BYE		( -- )
;		Exit eForth.

		$CODE	3,'BYE',BYE
		push	0			; error code
		OSCALL _ExitJob

;   ?RX		( -- c T | F )
;		Return input character and true, or a false if no input.

		$CODE	3,'?RX',QRX
		push	ebp			; save Forth machine
		push	esi
		OSCALL	_ReadKbd
		pop	esi			; restore Forth
		pop	ebp

		push	eax
		push	TRUEE

		cmp	eax, 0ffffffffh		; not cmp eax, 0ffh ???
		jne	short a_char		; not EOF -> a char 
		pop	eax			; undo T flag
		pop	eax			; remove char
		push	FALSEE			; false flag
a_char:		$NEXT

;   TX!		( c -- )
;		Send character c to the output device.

		$CODE	3,'TX!',TXSTO
		pop	ebx
		push	ebp			; save Forth
		push	esi
		push	ebx			; the character
		OSCALL	_TTYOut
		pop 	ebx			; remove character
		pop	esi			; restore Forth
		pop	ebp
		$NEXT

;   TYPE	( c-addr u -- )
;		Output u characters from c-addr.

		$CODE	4,'TYPE',TYPES
		pop	eax			; count
		pop	ebx			; addr
		push	ebp			; save Forth
		push	esi
		push	ebx			; the address
		push 	eax			; the count
		OSCALL	_Type
		pop	eax			; remove count
		pop 	ebx			; remove address
		pop	esi			; restore Forth
		pop	ebp
		$NEXT
		
;   SYSTEM	( c-addr u -- )
;		pass string to shell. Return code in RETURNCODE

		$CODE	6,'SYSTEM',SYSTEM
		pop	eax			; count
		pop	ebx			; address
		push	ebp			; save Forth
		push	esi
		push	ebx			; the address
		push 	eax			; the count
		OSCALL	_System
		mov	dword ptr [UPP+(RETCOD-UZERO)], eax
		pop	eax			; remove count
		pop 	ebx			; remove address
		pop	esi			; restore Forth
		pop	ebp
		$NEXT

		ALIGN
Febp:		DD	0
Fesi:		DD	0
Fcnt:		DD	0

;   SYSCALL	( [val]*n n #x -- return )
;		Call OS service x with n words on the stack.

		$CODE	7,'SYSCALL',SYSCALL
		pop	eax
		pop	ecx
		mov	dword ptr [Febp], ebp	; save Forth
		mov	dword ptr [Fesi], esi
		mov	dword ptr [Fcnt], ecx
		mov	ebx, dword ptr [JMPTABLE]
		call	[ebx+eax*4]
		mov	ecx, dword ptr [Fcnt]	; remove parameters
		mov	esi, dword ptr [Fesi]	; restore Forth
		mov	ebp, dword ptr [Febp]
		shl	ecx, 2
		add	esp, ecx
		push	eax			; return value
		$NEXT


;   !IO		( -- )
;		Initialize the serial I/O devices.

		$CODE	3,'!IO',STOIO
		$NEXT

;   CLS		( -- )
; 		Clear the screen

		$COLON	3,'CLS',CLRSCR
		DD 	DOLIT,0,DOLIT,8,SYSCALL,DROP,EXIT 

;   EOL		( -- )
; 		Clear to end of line

		$COLON	3,'EOL',CLREOL
		DD 	DOLIT,0,DOLIT,9,SYSCALL,DROP,EXIT 

;   EOS		( -- )
; 		Clear to end of screen

		$COLON	3,'EOS',CLREOS
		DD 	DOLIT,0,DOLIT,10,SYSCALL,DROP,EXIT 

;   BELL	( -- )
; 		Ring the terminal bell

		$COLON	4,'BELL',BELL
		DD 	DOLIT,0,DOLIT,11,SYSCALL,DROP,EXIT 

;   >BRIGHT	( -- )
; 		Bright attribute for text

		$COLON	7,'>BRIGHT',BRIGHT
		DD 	DOLIT,0,DOLIT,16,SYSCALL,DROP,EXIT 

;   >INVERSE	( -- )
; 		Inverse attribute for text

		$COLON	8,'>INVERSE',INVERSE
		DD 	DOLIT,0,DOLIT,17,SYSCALL,DROP,EXIT 

;   >BLINK	( -- )
; 		Blink attribute for text

		$COLON	6,'>BLINK',BLINK
		DD 	DOLIT,0,DOLIT,18,SYSCALL,DROP,EXIT 

;   DARK	( -- )
; 		Normal attribute for text

		$COLON	4,'DARK',DARK
		DD 	DOLIT,0,DOLIT,19,SYSCALL,DROP,EXIT 

;   AT-XY	( x y -- )
; 		Bring cursor to column x, row y

		$COLON	5,'AT-XY',ATXY
		DD 	DOLIT,2,DOLIT,20,SYSCALL,DROP,EXIT 

;   ?AT		( -- x y )
; 		Report present cursor position

		$COLON	3,'?AT',QAT
		DD 	DOLIT,0,DOLIT,21,SYSCALL
		DD	DUPP,DOLIT,16,RSHIFT
		DD	SWAP,DOLIT,0FFFFh,ANDD
		DD	EXIT 

;   ?MS	( -- x )
; 		Report current time in 1 ms ticks

		$COLON	3,'?MS',QMS
		DD 	DOLIT,0,DOLIT,22,SYSCALL
		DD	DOLIT,10,STAR
		DD	EXIT 

;  TIMER-RESET	( -- )
;		Reset the system timer

		$COLON	11,'TIMER-RESET',TIMRES
		DD 	QMS,T0,STORE,EXIT

;  .ELAPSED	( -- )
;		Print time elapsed since last call to TIMER-RESET

		$COLON	8,'.ELAPSED',DOTELAP
		DD	QMS,T0,AT,SUBB
		DD	BASE,AT,TOR,DECIM,UDOT,RFROM,BASE,STORE
		DD	DOTQP
		DB	11,' ms elapsed'
		DD	EXIT

;; The kernel

;   doLIT	( -- w )
;		Push an inline literal.

		$CODE	COMPO+5,'doLIT',DOLIT
		lodsd
		push	eax
		$NEXT

;   doDOIT	( a -- )
;		do what DOES> does

		$CODE	COMPO+6,'doDOIT',DODOIT
		pop	eax
		push	esi
		mov	esi, eax
		$NEXT

;   doLIST	( a -- )
;		Process colon list.

		$CODE	COMPO+6,'doLIST',DOLST
		sub	ebp, CELLL
		mov	[ebp], esi	; save on return stack
		pop	esi		; new list address
		$NEXT

;   next	( -- )
;		Run time code for the single index loop.
;		: next ( -- ) \ hilevel model
;		  r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

		$CODE	COMPO+4,'next',DONXT
		sub	dword ptr [ebp],1	; decrement the index
		jc	short NEXT1		; ?decrement below 0
		mov	esi, [esi+0]		; no, branch back again
		$NEXT
NEXT1:		add	ebp, CELLL		; yes, pop the index
		add	esi, CELLL		; continue past branch offset
		$NEXT

;   loop	( -- )
;		Run time code for the double index loop.

		$CODE	COMPO+4,'loop',DOLP
		mov	eax, dword ptr [ebp]	; increment the index
		inc	eax
		cmp	dword ptr 4[ebp], eax
		jbe	short LOOP1
		mov	dword ptr [ebp], eax
		mov	esi, [esi+0]		; no, branch back again
		$NEXT
LOOP1:		add	ebp, CELLL*2		; yes, pop the index+limit
		add	esi, CELLL		; continue past branch offset
		$NEXT

;   +loop	( -- )
;		Run time code for the double index +loop.

		$CODE	COMPO+5,'+loop',DOPLP
		mov	eax, dword ptr [ebp]	; increment the index
		pop	ebx
		add	eax, ebx
		cmp	dword ptr 4[ebp], eax
		jbe	short PLOOP1
		mov	dword ptr [ebp], eax
		mov	esi, [esi+0]		; no, branch back again
		$NEXT
PLOOP1:		add	ebp, CELLL*2		; yes, pop the index+limit
		add	esi, CELLL		; continue past branch offset
		$NEXT

;   dup+loop	( -- )
;		Run time code for the double index dup+loop.

		$CODE	COMPO+8,'dup+loop',DODPLP
		mov	eax, dword ptr [ebp]	; increment the index
		add	eax, [esp]
		cmp	dword ptr 4[ebp], eax
		jbe	short DPLOOP1
		mov	dword ptr [ebp], eax
		mov	esi, [esi+0]		; no, branch back again
		$NEXT
DPLOOP1:	add	ebp, CELLL*2		; yes, pop the index+limit
		add	esi, CELLL		; continue past branch offset
		$NEXT

;   ?branch	( f -- )
;		Branch if flag is zero.

		$CODE	COMPO+7,'?branch',QBRAN
		pop	ebx			; pop flag
		or	ebx, ebx		; ?flag=0
		jz	short BRAN1		; yes, so branch
		add	esi, CELLL		; point IP to next cell
		$NEXT
BRAN1:		mov	esi,[esi]		; ip:=(ip)
		$NEXT

;   doQBRAN	( -- )
		$NONAME	doQBRAN
		DD	DOLIT,QBRAN,COMPIC,EXIT 

;   branch	( -- )
;		Branch to an inline address.

		$CODE	COMPO+6,'branch',BRAN
		mov	esi,[esi]		; IP:=(IP)
		$NEXT

;   ?dup_0=_if	( f -- )
;		test and branch if flag is not zero.

		$CODE	COMPO+10,'?dup_0=_if',QDZEQIF
		pop	ebx			; pop flag
		or	ebx, ebx		; ?flag=0
		jnz	short BRAN2		; no, so branch
		add	esi, CELLL		; point IP to next cell
		$NEXT
BRAN2:		push	ebx
		mov	esi,[esi]		; ip:=(ip)
		$NEXT

;   0=_if	( f -- )
;		branch if flag is not zero.

		$CODE	COMPO+5,'0=_if',ZEQIF
		pop	ebx			; pop flag
		or	ebx, ebx		; ?flag=0
		jnz	short BRAN3		; no, so branch
		add	esi, CELLL		; point IP to next cell
		$NEXT
BRAN3:		mov	esi,[esi]		; ip:=(ip)
		$NEXT

;   dup_if	( f -- )
;		test and branch if flag is zero.

		$CODE	COMPO+6,'dup_if',DIFF
		pop	ebx			; pop flag
		push	ebx
		or	ebx, ebx		; ?flag=0
		jz	short BRAN4		; yes, so branch
		add	esi, CELLL		; point IP to next cell
		$NEXT
BRAN4:		mov	esi,[esi]		; ip:=(ip)
		$NEXT


;   EXECUTE	( ca -- )
;		Execute the word at ca.

		$CODE	7,'EXECUTE',EXECU
		pop	ebx
		jmp	ebx			; jump to the code address

;   EXIT	( -- )
;		Terminate a colon definition.

		$CODE	4,'EXIT',EXIT
		mov	esi, [ebp]		; pop return address
		add	ebp, CELLL		; adjust RP
		$NEXT

;		Terminate a colon definition.
;		Same as EXIT, but this helps SEE to stop gracefully.

		$NONAME	EXITT
		mov	esi, [ebp]		; pop return address
		add	ebp, CELLL		; adjust RP
		$NEXT

;   C!		( c b -- )
;		Pop the data stack to byte memory.

		$CODE	2,'C!',CSTOR
		pop	ebx
		pop	eax
		mov	byte ptr [ebx], al
		$NEXT

;   C@		( b -- c )
;		Push byte memory location to the data stack.

		$CODE	2,'C@',CAT
		pop	ebx
		xor	eax, eax		; EAX=0 zero the hi byte
		mov	al, [ebx]
		push	eax
		$NEXT

;   !		( w a -- )
;		Pop the data stack to memory.

		$CODE	1,'!',STORE
		pop	ebx
		pop	dword ptr [ebx]
		$NEXT

;   @		( a -- w )
;		Push memory location to the data stack.

		$CODE	1,'@',AT
		pop	ebx
		push	dword ptr [ebx]
		$NEXT

;   nC!		( w -- )
;		Pop the data stack to inline address.

		$CODE	3,'nC!',NCSTORE
		pop	ebx
		lodsd
		mov	byte ptr [eax],bl
		$NEXT

;   nC@		( -- w )
;		Push inline memory location content.

		$CODE	3,'nC@',NCAT
		lodsd
		xor	ebx,ebx
		mov	bl,[eax]
		push	ebx
		$NEXT

;   I!		( w -- )
;		Pop the data stack to address on top of return stack.

		$CODE	2,'I!',ISTORE
		mov	ebx, dword ptr [ebp]
		pop	dword ptr [ebx]
		$NEXT

;   I@		( -- w )
;		Push the data in the address on top of the return stack.

		$CODE	2,'I@',IAT
		mov	ebx, dword ptr [ebp]
		push	dword ptr [ebx]
		$NEXT

;   IC!		( w -- )
;		Pop the data stack to address on top of return stack.

		$CODE	3,'IC!',ICSTORE
		mov	ebx, dword ptr [ebp]
		pop	eax
		mov	byte ptr [ebx], al
		$NEXT

;   IC@		( -- w )
;		Push the data in the address on top of the return stack.

		$CODE	3,'IC@',ICAT
		mov	ebx, dword ptr [ebp]
		movzx	eax, byte ptr [ebx]
		push	eax
		$NEXT

;   I2!		( d -- )
;		Pop the data stack to address on top of return stack.

		$CODE	3,'I2!',IDSTORE
		mov	ebx, dword ptr [ebp]
		pop	dword ptr  [ebx]
		pop	dword ptr 4[ebx]
		$NEXT

;   I2@		( -- d )
;		Push the data in the address on top of the return stack.

		$CODE	3,'I2@',IDAT
		mov	ebx, dword ptr [ebp]
		push	dword ptr 4[ebx]
		push	dword ptr  [ebx]
		$NEXT

;   nIC!	( -- )
;		Pop inline data to address on top return stack.

		$CODE	4,'nIC!',NICSTORE
		mov	ebx, dword ptr [ebp]
		lodsd
		mov	byte ptr [ebx],al
		$NEXT

;   n!		( w -- )
;		Pop the data stack to inline address.

		$CODE	2,'n!',NSTORE
		pop	ebx
		lodsd
		mov	dword ptr [eax],ebx
		$NEXT

;   nI!		( -- )
;		Pop inline data to address on top return stack.

		$CODE	3,'nI!',NISTORE
		mov	ebx, dword ptr [ebp]
		lodsd
		mov	[ebx],eax
		$NEXT

;   n@		( a -- w )
;		Push inline memory location content.

		$CODE	2,'n@',NAT
		lodsd
		push	dword ptr [eax]
		$NEXT

;   over_@	( n1 n2 -- n1 n2 w )
;		Push memory location content pointed to by SEC.

		$CODE	6,'over_@',OVERAT
		mov	eax,4[esp]
		push	dword ptr [eax]
		$NEXT

;   n2!		( d -- )
;		Store the double integer to inline address.

		$CODE	3,'n2!',NDSTOR
		lodsd
		pop	dword ptr 0[eax]
		pop	dword ptr 4[eax]
		$NEXT

;   n2@		( -- d )
;		Fetch double integer from inline address.

		$CODE	3,'n2@',NDAT
		lodsd
		push	dword ptr 4[eax]
		push	dword ptr 0[eax]
		$NEXT

;   2!		( d a -- )
;		Store the double integer to address a.

		$CODE	2,'2!',DSTOR
		pop	ebx
		pop	dword ptr 0[ebx]
		pop	dword ptr 4[ebx]
		$NEXT

;   2@		( a -- d )
;		Fetch double integer from address a.

		$CODE	2,'2@',DAT
		pop	ebx
		push	dword ptr 4[ebx]
		push	dword ptr 0[ebx]
		$NEXT

;   RP@		( -- a )
;		Push the current RP to the data stack.

		$CODE	3,'RP@',RPAT
		push	ebp
		$NEXT

;   RP!		( a -- )
;		Set the return stack pointer.

		$CODE	COMPO+3,'RP!',RPSTO
		pop	ebp
		$NEXT

;   R>		( -- w )
;		Pop the return stack to the data stack.

		$CODE	COMPO+2,'R>',RFROM
		push	dword ptr [ebp]
		add	ebp, CELLL		; adjust RP
		$NEXT


;   r>_+_>r	( n -- )
;		Add TOS to top of return stack.

		$CODE	COMPO+7,'r>_+_>r',RFROMPTOR
		pop	ebx
		add	[ebp],ebx
		$NEXT

;   R@		( -- w )
;		Copy top of return stack to the data stack.

		$CODE	2,'R@',RAT
		push	dword ptr [ebp]
		$NEXT

;   I		( -- w )
;		Copy loop index to the data stack.

		$CODE	1,'I',II
		push	dword ptr [ebp]
		$NEXT

;   J		( -- w )
;		Copy second loop index to the data stack.

		$CODE	1,'J',JJ
		push	dword ptr 8[ebp]
		$NEXT

;   K		( -- w )
;		Copy third loop index to the data stack.

		$CODE	1,'K',KK
		push	dword ptr 16[ebp]
		$NEXT

;   LEAVE	( -- w )
;		Leave loop at next possibility. (Not standard!)

		$CODE	5,'LEAVE',LEAVEE
		mov	eax, dword ptr 4[ebp]
		mov	dword ptr [ebp],eax
		$NEXT

;   >R		( w -- )
;		Push the data stack to the return stack.

		$CODE	COMPO+2,'>R',TOR
		sub	ebp, CELLL		; adjust RP
		pop	dword ptr [ebp]		; push
		$NEXT

;   SP@		( -- a )
;		Push the current data stack pointer.

		$CODE	3,'SP@',SPAT
		mov	ebx,esp			; BX indexes the data stack
		push	ebx
		$NEXT

;   SP!		( a -- )
;		Set the data stack pointer.

		$CODE	3,'SP!',SPSTO
		pop	esp
		$NEXT

;   DROP	( w -- )
;		Discard top stack item.

		$CODE	4,'DROP',DROP
		add	esp, CELLL		; adjust SP
		$NEXT

;   DUP		( w -- w w )
;		Duplicate the top stack item.

		$CODE	3,'DUP',DUPP
		mov	ebx, esp		; BX indexes the data stack
		push	dword ptr [ebx]
		$NEXT

;   SWAP	( w1 w2 -- w2 w1 )
;		Exchange top two stack items.

		$CODE	4,'SWAP',SWAP
		pop	ebx
		pop	eax
		push	ebx
		push	eax
		$NEXT

;   OVER	( w1 w2 -- w1 w2 w1 )
;		Copy second stack item to top.

		$CODE	4,'OVER',OVER
		mov	ebx, esp		; BX indexes the stack
		push	dword ptr [ebx+CELLL]
		$NEXT

;   0<		( n -- t )
;		Return true if n is negative.

		$CODE	2,'0<',ZLESS
		pop	eax
		cdq				; sign extend
		push	edx
		$NEXT

;   0=		( n -- t )
;		Return true if n is 0.

		$CODE	2,'0=',ZEQ
		pop	eax
		or	eax, eax
		jz	SHORT iszero
		push	FALSEE
		$NEXT
iszero:		push	TRUEE
		$NEXT

;   AND		( w w -- w )
;		Bitwise AND.

		$CODE	3,'AND',ANDD
		pop	ebx
		pop	eax
		and	ebx, eax
		push	ebx
		$NEXT

;   OR		( w w -- w )
;		Bitwise inclusive OR.

		$CODE	2,'OR',ORR
		pop	ebx
		pop	eax
		or	ebx, eax
		push	ebx
		$NEXT

;   XOR		( w w -- w )
;		Bitwise exclusive OR.

		$CODE	3,'XOR',XORR
		pop	ebx
		pop	eax
		xor	ebx, eax
		push	ebx
		$NEXT

;   nAND	( w -- w )
;		Bitwise AND with inline literal.

		$CODE	4,'nAND',NANDD
		pop	ebx
		lodsd
		and	ebx, eax
		push	ebx
		$NEXT

;   nOR		( w -- w )
;		Bitwise inclusive OR with inline literal.

		$CODE	3,'nOR',NORR
		pop	ebx
		lodsd
		or	ebx, eax
		push	ebx
		$NEXT

;   nXOR	( w -- w )
;		Bitwise exclusive OR with inline literal.

		$CODE	4,'nXOR',NXORR
		pop	ebx
		lodsd
		xor	ebx, eax
		push	ebx
		$NEXT

;   RSHIFT	( w1 w2 -- w3 )
;		Logically shift w1 right over w2 bit positions.

		$CODE	6,'RSHIFT',RSHIFT
		pop	ecx
		pop	eax
		shr	eax, cl
		push	eax
		$NEXT

;   LSHIFT	( w1 w2 -- w3 )
;		Logically shift w1 left over w2 bit positions.

		$CODE	6,'LSHIFT',LSHIFT
		pop	ecx
		pop	eax
		shl	eax, cl
		push	eax
		$NEXT

;   UM+		( u u -- udsum )
;		Add two unsigned single numbers and return a double sum.

		$CODE	3,'UM+',UPLUS
		xor	ecx, ecx		; ECX=0 initial carry flag
		pop	ebx
		pop	eax
		add	eax, ebx
		rcl	ecx, 1			; get carry
		push	eax			; push sum
		push	ecx			; push carry
		$NEXT


;   ROT		( w1 w2 w3 -- w2 w3 w1 )
;		Rot 3rd item to top.

		$CODE	3,'ROT',ROT
		pop	ecx
		pop	ebx
		pop	eax
		push	ebx
		push	ecx
		push	eax
		$NEXT

;   2DROP	( w w -- )
;		Discard two items on stack.

		$CODE	5,'2DROP',DDROP
		pop	eax
		pop	eax
		$NEXT

;   2DUP	( w1 w2 -- w1 w2 w1 w2 )
;		Duplicate top two items.

		$CODE	4,'2DUP',DDUP
		pop	ebx
		pop	eax
		push	eax
		push	ebx
		push	eax
		push	ebx
		$NEXT

;   +		( w w -- sum )
;		Add top two items.

		$CODE	1,'+',PLUS
		pop	ebx
		pop	eax
		add	eax, ebx
		push	eax
		$NEXT

;   I+		( n -- w )
;		Add n to the number on top of the return stack.

		$CODE	2,'I+',IPLUS
		mov	ebx, dword ptr [ebp]
		add	[esp], ebx
		$NEXT

;   I-		( n -- w )
;		Subtract the number on top of the return stack from n.

		$CODE	2,'I-',IMINUS
		mov	ebx, dword ptr [ebp]
		sub	[esp], ebx
		$NEXT

;   swap_n+_swap ( n1 n2 -- n1+n n2 )
;		Add inline literal to SEC.

		$CODE	12,'swap_n+_swap',SWNPLUSSW
		lodsd
		add	4[esp],eax
		$NEXT

;   n+		( w -- sum )
;		Add top with inline literal

		$CODE	2,'n+',NPLUS
		pop	ebx
		lodsd
		add	eax, ebx
		push	eax
		$NEXT

;   1+		( w -- sum )
;		Increment top item.

		$CODE	2,'1+',ONEPLUS
		pop	eax
		inc	eax
		push	eax
		$NEXT

;   2+		( w w -- sum )
;		Add 2 to top item.

		$CODE	2,'2+',TWOPLUS
		pop	eax
		inc	eax
		inc	eax
		push	eax
		$NEXT

;   2*		( w -- sum )
;		Multiply TOS by 2

		$CODE	2,'2*',TWOSTAR
		pop	eax
		add	eax, eax
		push	eax
		$NEXT

;   2/		( w1 -- w2 )
;		Arithmetically shift w1 right once.

		$CODE	2,'2/',TWOSLASH
		pop	eax
		sar	eax, 1
		push	eax
		$NEXT

;   NOT		( w -- w )
;		One's complement of tos.

		$CODE	3,'NOT',INVER
		pop	eax
		not	eax
		push	eax
		$NEXT

;   NEGATE	( n -- -n )
;		Two's complement of tos.

		$CODE	6,'NEGATE',NEGAT
		pop	eax
		neg	eax	
		push	eax
		$NEXT

;   -		( n1 n2 -- n1-n2 )
;		Subtraction.

		$CODE	1,'-',SUBB
		pop	ebx
		pop	eax
		sub	eax, ebx
		push	eax
		$NEXT

;   n-		( n -- w )
;		Subtract inline literal from top.

		$CODE	2,'n-',NSUBB
		pop	ebx
		lodsd
		sub	ebx, eax
		push	ebx
		$NEXT

;   1-		( w -- sum )
;		Decrement top item.

		$CODE	2,'1-',ONEMINUS
		pop	eax
		dec	eax
		push	eax
		$NEXT

;   2-		( w w -- sum )
;		Subtract two from top item.

		$CODE	2,'2-',TWOMINUS
		pop	eax
		dec	eax
		dec	eax
		push	eax
		$NEXT

;   =		( w w -- t )
;		Return true if top two are equal.

		$CODE	1,'=',EQUAL
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		je	SHORT EQU1
		push	FALSEE
		$NEXT
EQU1:		push	TRUEE
		$NEXT

;   <>		( w w -- t )
;		Return true if top two are not equal.

		$CODE	2,'<>',NEQUAL
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jne	SHORT NEQU1
		push	FALSEE
		$NEXT
NEQU1:		push	TRUEE
		$NEXT

;   U<		( u u -- t )
;		Unsigned compare of top two items.

		$CODE	2,'U<',ULESS
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jb	SHORT ULES1
		push	FALSEE
		$NEXT
ULES1:		push	TRUEE
		$NEXT

;   U>		( u u -- t )
;		Unsigned compare of top two items.

		$CODE	2,'U>',UGREATER
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		ja	SHORT UGREAT1
		push	FALSEE
		$NEXT
UGREAT1:	push	TRUEE
		$NEXT

;   <		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE	1,'<',LESS
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jl	SHORT LESS1
		push	FALSEE
		$NEXT
LESS1:		push	TRUEE
		$NEXT

;  n<		( n2 -- t )
;		Signed compare of top with inline literal

		$CODE	COMPO+2,'n<',NLESS
		pop	ebx	
		lodsd
		cmp	ebx,eax
		jl	SHORT NLESS1
		push	FALSEE
		$NEXT
NLESS1:		push	TRUEE
		$NEXT

;  dup_n<	( n2 -- n2 t )
;		Compare top with inline literal, non-destructively

		$CODE	COMPO+6,'dup_n<',DUPNLESS
		mov	ebx,[esp]
		lodsd
		cmp	ebx, eax
		jl	SHORT DNLESS1
		push	FALSEE
		$NEXT
DNLESS1:	push	TRUEE
		$NEXT

;   >		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE	1,'>',GREATER
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jg	SHORT GREAT1
		push	FALSEE
		$NEXT
GREAT1:		push	TRUEE
		$NEXT


;   <=		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE	2,'<=',LESSE
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jle	SHORT LESSE1
		push	FALSEE
		$NEXT
LESSE1:		push	TRUEE
		$NEXT

;   >=		( n1 n2 -- t )
;		Signed compare of top two items.

		$CODE	2,'>=',GREATERE
		pop	ebx	
		pop	eax
		cmp	eax,ebx
		jge	SHORT GREATE1
		push	FALSEE
		$NEXT
GREATE1:	push	TRUEE
		$NEXT


;  n=		( n2 -- t )
;		Compare top with inline literal

		$CODE	COMPO+2,'n=',NEQUALL
		pop	ebx	
		lodsd
		cmp	eax,ebx
		je	SHORT NEQUALL1
		push	FALSEE
		$NEXT
NEQUALL1:	push	TRUEE
		$NEXT

;  dup_n=	( n2 -- n2 t )
;		Compare top with inline literal, non-destructively

		$CODE	COMPO+6,'dup_n=',DUPNEQUALL
		mov	ebx,[esp]
		lodsd
		cmp	eax,ebx
		je	SHORT DNEQUALL1
		push	FALSEE
		$NEXT
DNEQUALL1:	push	TRUEE
		$NEXT

;  drop_n	( n1 -- n2 )
;		Replace TOS with inline literal.

		$CODE	COMPO+6,'drop_n',DROPN
		lodsd
		mov	[esp],eax
		$NEXT

;   dup_n-	( n -- n w )
;		Subtract inline literal from duplicated TOS.

		$CODE	6,'dup_n-',DNSUBB
		mov	ebx,[esp]
		lodsd
		sub	ebx, eax
		push	ebx
		$NEXT


;   swap_n-	( n1 n2 -- n2 n1-w )
;		Swap and subtract inline literal from new TOS.

		$CODE	7,'swap_n-',SWNSUBB
		pop	eax
		pop	ebx
		push	eax
		lodsd
		sub	ebx, eax
		push	ebx
		$NEXT

;   swap_n+	( n1 n2 -- n2 n1+w )
;		Swap and add inline literal to new TOS.

		$CODE	7,'swap_n+',SWNADD
		pop	eax
		pop	ebx
		push	eax
		lodsd
		add	ebx, eax
		push	ebx
		$NEXT

;; Divide

;   UM/MOD	( udl udh un -- ur uq )
;		Unsigned divide of double by single. Return mod and quotient.

		$CODE	6,'UM/MOD',UMMOD
		pop	ecx
		pop	edx
		pop	eax
		div	ecx
		push	edx
		push	eax
		$NEXT

;   M/MOD	( d n -- r q )
;		Signed divide of double by single. Return mod and quotient.

		$CODE	5,'M/MOD',MSMOD
		pop	ecx
		pop	edx
		pop	eax
		idiv	ecx
		push	edx
		push	eax
		$NEXT

;   /MOD	( n n -- r q )
;		Signed divide. Return mod and quotient.

		$CODE	4,'/MOD',SLMOD
		pop	ecx
		pop	eax
		cdq
		idiv	ecx
		push	edx
		push	eax
		$NEXT

;   MOD		( n n -- r )
;		Signed divide. Return mod only.

		$CODE	3,'MOD',MODD
		pop	ecx
		pop	eax
		cdq
		idiv	ecx
		push	edx
		$NEXT

;   /		( n n -- q )
;		Signed divide. Return quotient only.

		$CODE	1,'/',SLASH
		pop	ecx
		pop	eax
		cdq
		idiv	ecx
		push	eax
		$NEXT

;; Multiply

;   UM*		( u u -- ud )
;		Unsigned multiply. Return double product.

		$CODE	3,'UM*',UMSTA
		pop	ebx
		pop	eax
		mul	ebx	
		push	eax
		push	edx
		$NEXT

;   *		( n n -- n )
;		Signed multiply. Return single product.

		$CODE	1,'*',STAR
		pop	ebx
		pop	eax
		imul	ebx	
		push	eax
		$NEXT

;   n*		( n -- n )
;		Signed multiply of top with inline literal.

		$CODE	2,'n*',NSTAR
		pop	ebx
		lodsd
		imul	ebx	
		push	eax
		$NEXT

;   M*		( n n -- d )
;		Signed multiply. Return double product.

		$CODE	2,'M*',MSTAR
		pop	ebx
		pop	eax
		imul	ebx	
		push	eax
		push	edx
		$NEXT

;   */MOD	( n1 n2 n3 -- r q )
;		Multiply n1 and n2, then divide by n3. Return mod and quotient.

		$CODE	5,'*/MOD',SSMOD
		pop	ecx
		pop	ebx
		pop	eax
		imul	ebx
		idiv	ecx
		push	edx
		push	eax
		$NEXT

;   */		( n1 n2 n3 -- q )
;		Multiply n1 by n2, then divide by n3. Return quotient only.

		$CODE	2,'*/',STASL
		pop	ecx
		pop	ebx
		pop	eax
		imul	ebx
		idiv	ecx
		push	eax
		$NEXT

;; System and user variables

;   doVAR	( -- a )
;		Run time routine for dumb VARIABLE and CREATE.

		$COLON	COMPO+5,'doVAR',DOVAR
		DD	RFROM,EXIT

;   		( -- a )
;		Run time routine for smart VARIABLE.

		$NONAME	SDOVAR
		DD	RFROM
		DD	STATE,AT,QBRAN,notcompv
		DD	DOLIT,DOLIT,COMPIC,COMMA
notcompv:	DD	EXIT


;   		( -- n )
;		Run time routine for smart CONSTANT

		$NONAME	SDOCONS
		DD	RFROM,AT
		DD	STATE,AT,QBRAN,notcompc
		DD	DOLIT,DOLIT,COMPIC,COMMA
notcompc:	DD	EXIT


;   UP		( -- a )
;		Pointer to the user area.

		$COLON	2,'UP',UP
		DD	DOVAR
		DD	UPP

;   doUSER	( -- a )
;		Run time routine for user variables.

		$COLON	COMPO+6,'doUSER',DOUSE
		DD	RFROM,AT,UP,AT,PLUS,EXIT

;   SP0		( -- a )
;		Pointer to bottom of the data stack.

		$USER	3,'SP0',SZERO

;   RP0		( -- a )
;		Pointer to bottom of the return stack.

		$USER	3,'RP0',RZERO

;   '?KEY	( -- a )
;		Execution vector of ?KEY.

		$USER	5,"'?KEY",TQKEY

;   'EMIT	( -- a )
;		Execution vector of EMIT.

		$USER	5,"'EMIT",TEMIT

;   'EXPECT	( -- a )
;		Execution vector of EXPECT.

		$USER	7,"'EXPECT",TEXPE

;   'TAP	( -- a )
;		Execution vector of TAP.

		$USER	4,"'TAP",TTAP

;   'ECHO	( -- a )
;		Execution vector of ECHO.

		$USER	5,"'ECHO",TECHO

;   'PROMPT	( -- a )
;		Execution vector of PROMPT.

		$USER	7,"'PROMPT",TPROM

;   BASE	( -- a )
;		Storage of the radix base for numeric I/O.

		$USER	4,'BASE',BASE

;   tmp		( -- a )
;		A temporary storage location used in parse and find.

		$USER	COMPO+3,'tmp',TEMP

;   SPAN	( -- a )
;		Hold character count received by EXPECT.

		$USER	4,'SPAN',SPAN

;   >IN		( -- a )
;		Hold the character pointer while parsing input stream.

		$USER	3,'>IN',INN

;   #TIB	( -- a )
;		Hold the current count in and address of the terminal 
;		input buffer.

		$USER	4,'#TIB',NTIB

; <nameless>
;		hold the base address of the terminal input buffer
		_USER = _USER+CELLL		


;   CSP		( -- a )
;		Hold the stack pointer for error checking.

		$USER	3,'CSP',CSP

;   'EVAL	( -- a )
;		Execution vector of EVAL.

		$USER	5,"'EVAL",TEVAL

;   'NUMBER	( -- a )
;		Execution vector of NUMBER?.

		$USER	7,"'NUMBER",TNUMB

;   HLD		( -- a )
;		Hold a pointer in building a numeric output string.

		$USER	3,'HLD',HLD

;   HANDLER	( -- a )
;		Hold the return stack pointer for error handling.

		$USER	7,'HANDLER',HANDL

;   RETURNCODE	( -- a )
;		What SYSTEM returns 

		$USER	10,'RETURNCODE',RETURNCODE

;   T0		( -- a )
;		Used by .ELAPSED

		$USER	2,'T0',T0

;   last	( -- a )
;		Used by optimizer

		$USER	4,'last',lastt

;   penult	( -- a )
;		Used by optimizer

		$USER	6,'penult',penult

;   STATE	( -- a )
;		TRUE if compiling, FALSE if interpreting.

		$USER	5,'STATE',STATE

;   (?BRANCH)	( -- a )
;		Vector for ?branch,

		$USER	9,'(?BRANCH)',bqbran

;   CONTEXT	( -- a )
;		An area to specify vocabulary search order.

		$USER	7,'CONTEXT',CNTXT
		_USER = _USER+VOCSS*CELLL	; vocabulary stack

;   CURRENT	( -- a )
;		Point to the vocabulary to be extended.

		$USER	7,'CURRENT',CRRNT
		_USER = _USER+CELLL		; vocabulary link pointer

;   CP		( -- a )
;		Point to the top of the code dictionary.

		$USER	2,'CP',CP

;   NP		( -- a )
;		Point to the bottom of the name dictionary.

		$USER	2,'NP',NP

;   LAST	( -- a )
;		Point to the last name in the name dictionary.

		$USER	4,'LAST',LAST

;; Common functions

;   doVOC	( -- )
;		Run time action of VOCABULARY's.

		$COLON	COMPO+5,'doVOC',DOVOC
		DD	RFROM,CNTXT,STORE,EXIT

;   FORTH	( -- )
;		Make FORTH the context vocabulary.

		$COLON	5,'FORTH',FORTH
		DD	DOVOC
		DD	0			; vocabulary head pointer
		DD	0			; vocabulary link pointer

;   DNEGATE	( d -- -d )
;		Two's complement of top double.

		$CODE	7,'DNEGATE',DNEGA
		pop	eax
		not	eax
		pop	ecx
		not	ecx
		add	ecx,1
		adc	eax,0
		push	ecx
		push	eax
		$NEXT

;   ABS		( n -- n )
;		Return the absolute value of n.

		$CODE	3,'ABS',ABSS
		pop	ebx
		or	ebx,ebx
		jns	SHORT ABS1
		neg	ebx
ABS1:		push	ebx
		$NEXT

;   ?DUP	( w -- w w | 0 )
;		Dup tos if it is not zero.

		$CODE	4,'?DUP',QDUP
		pop	ebx
		or	ebx,ebx
		jz	QDUP1
		push	ebx
QDUP1:		push	ebx
		$NEXT

;   MAX		( n n -- n )
;		Return the greater of two top stack items.

		$CODE	3,'MAX',MAX
		pop	ebx
		pop	eax
		cmp	eax,ebx		; eax-ebx
		jg	SHORT MAX1
		push	ebx
		$NEXT	
MAX1:		push	eax
		$NEXT

;   MIN		( n n -- n )
;		Return the smaller of top two stack items.

		$CODE	3,'MIN',MIN
		pop	ebx
		pop	eax
		cmp	eax,ebx		; eax-ebx
		jg	SHORT MIN1
		push	eax
		$NEXT	
MIN1:		push	ebx
		$NEXT

;   WITHIN	( u ul uh -- t )
;		Return true if u is within the range of ul and uh. 
;		( ul <= u < uh )

		$COLON	6,'WITHIN',WITHI
		DD	OVER,SUBB,TOR
		DD	SUBB,RFROM,ULESS,EXIT

;; Miscellaneous

;   CELL+	( a -- a )
;		Add cell size in byte to address.

		$CODE	5,'CELL+',CELLP
		pop	eax
		add	eax, CELLL
		push	eax
		$NEXT

;   CELL-	( a -- a )
;		Subtract cell size in byte from address.

		$CODE	5,'CELL-',CELLM
		pop	eax
		sub	eax, CELLL
		push	eax
		$NEXT

;   CELLS	( n -- n )
;		Multiply tos by cell size in bytes.

		$CODE	5,'CELLS',CELLS
		pop	eax
		shl	eax,2
		push	eax
		$NEXT

;   ALIGNED	( b -- a )
;		Align address to the cell boundary.

		$COLON	7,'ALIGNED',ALGND
		DD	DUPP,DOLIT,0,DOLIT,CELLL
		DD	UMMOD,DROP,DUPP
		DD	QBRAN,ALGN1
		DD	DOLIT,CELLL,SWAP,SUBB
ALGN1:		DD	PLUS,EXIT

;   ALIGN	( -- )
;		Align dictionary

		$COLON	5,'ALIGN',ALGN
		DD	HERE,ALGND,CP,STORE,EXIT

;  		( -- 32 )
;		Return 32, the blank character (internal use).

		$NONAME	BLANK
		DD	DOLIT,' ',EXIT
		
;   BL		( -- 32 )
;		Return 32, the blank character.

		$COLON	2,'BL',IBLANK
		DD	DOLIT,' '
		DD	STATE,AT,QBRAN,nbl
		DD	DOLIT,DOLIT,COMPIC,COMMA
nbl:		DD	EXIT

;   >CHAR	( c -- c )
;		Filter non-printing characters.

		$COLON	5,'>CHAR',TCHAR
		DD	DOLIT,07FH,ANDD,DUPP	; mask msb
		DD	DOLIT,127,BLANK,WITHI	; check for printable
		DD	QBRAN,TCHA1
		DD	DROP,DOLIT,'_'		; replace non-printables
TCHA1:		DD	EXIT

;   DEPTH	( -- n )
;		Return the depth of the data stack.

		$COLON	5,'DEPTH',DEPTH
		DD	SPAT,SZERO,AT,SWAP,SUBB
		DD	DOLIT,CELLL,SLASH,EXIT

;   PICK	( ... +n -- ... w )
;		Copy the nth stack item to tos.

		$COLON	4,'PICK',PICK
		DD	ONEPLUS,CELLS
		DD	SPAT,PLUS,AT,EXIT

;; Memory access

;   +!		( n a -- )
;		Add n to the contents at address a.

		$CODE	2,'+!',PSTOR
		pop	ebx
		pop	eax
		add	[ebx],eax
		$NEXT
	
;   COUNT	( b -- b +n )
;		Return count byte of a string and add 1 to byte address.

		$CODE	5,'COUNT',COUNT
		pop	ebx	
		movzx	eax,byte ptr [ebx]
		inc	ebx
		push	ebx
		push	eax
		$NEXT

;   HERE	( -- a )
;		Return the top of the code dictionary.

		$COLON	4,'HERE',HERE
		DD	CP,AT,EXIT

;   PAD		( -- a )
;		Return the address of the text buffer above the code 
;		dictionary.

		$COLON	3,'PAD',PAD
		DD	HERE,DOLIT,128,PLUS,EXIT

;   TIB		( -- a )
;		Return the address of the terminal input buffer.

		$COLON	3,'TIB',TIB
		DD	NTIB,CELLP,AT,EXIT

;   @EXECUTE	( a -- )
;		Execute vector stored in address a.

		$COLON	8,'@EXECUTE',ATEXE
		DD	AT,QDUP			; ?address or zero
		DD	QBRAN,EXE1
		DD	EXECU			; execute if non-zero
EXE1:		DD	EXIT			; do nothing if zero

;   CMOVE	( b1 b2 u -- )
;		Copy u bytes from b1 to b2.

		$COLON	5,'CMOVE',CMOVE
		DD	TOR
		DD	BRAN,CMOV2
CMOV1:		DD	TOR,DUPP,CAT
		DD	RAT,CSTOR
		DD	ONEPLUS
		DD	RFROM,ONEPLUS
CMOV2:		DD	DONXT,CMOV1
		DD	DDROP,EXIT


;   FILL	( b u c -- )
;		Fill u bytes of character c to area beginning at b.

		$CODE	4,'FILL',FILL

		pop	eax
		pop	ecx
		pop	ebx
		jecxz	SHORT READY
ANEW:		mov	[ebx],al
		inc	ebx
		dec	ecx	
		jnz	ANEW
READY:		$NEXT


;   -TRAILING	( b u -- b u )
;		Adjust the count to eliminate trailing white space.

		$COLON	9,'-TRAILING',DTRAI
		DD	TOR
		DD	BRAN,DTRA2
DTRA1:		DD	BLANK,OVER,RAT,PLUS,CAT,LESS
		DD	QBRAN,DTRA2
		DD	RFROM,ONEPLUS,EXIT
DTRA2:		DD	DONXT,DTRA1
		DD	DOLIT,0,EXIT

;   PACK$	( b u a -- a )
;		Build a counted string with u characters from b. Null fill.

		$COLON	5,'PACK$',PACKS
		DD	ALGND,DUPP,TOR		; strings only on cell boundary
		DD	OVER,DUPP,DOLIT,0
		DD	DOLIT,CELLL,UMMOD,DROP	; count mod cell (floored!)
		DD	SUBB,OVER,PLUS
		DD	DOLIT,0,SWAP,STORE	; null fill cell
		DD	DDUP,CSTOR,ONEPLUS	; save count
		DD	SWAP,CMOVE,RFROM,EXIT	; move string

;; Numeric output, single precision

;   DIGIT	( u -- c )
;		Convert digit u to a character.

		$COLON	5,'DIGIT',DIGIT
		DD	DOLIT,9,OVER,LESS
		DD	DOLIT,7,ANDD,PLUS
		DD	DOLIT,'0',PLUS,EXIT

;   EXTRACT	( n base -- n c )
;		Extract the least significant digit from n.

		$COLON	7,'EXTRACT',EXTRC
		DD	DOLIT,0,SWAP,UMMOD
		DD	SWAP,DIGIT,EXIT

;   <#		( -- )
;		Initiate the numeric output process.

		$COLON	2,'<#',BDIGS
		DD	PAD,HLD,STORE,EXIT

;   HOLD	( c -- )
;		Insert a character into the numeric output string.

		$COLON	4,'HOLD',HOLD
		DD	HLD,AT,ONEMINUS
		DD	DUPP,HLD,STORE,CSTOR,EXIT

;   #		( u -- u )
;		Extract one digit from u and append the digit to output string.

		$COLON	1,'#',DIG
		DD	BASE,AT,EXTRC,HOLD,EXIT

;   #S		( u -- 0 )
;		Convert u until all digits are added to the output string.

		$COLON	2,'#S',DIGS
DIGS1:		DD	DIG,DUPP
		DD	QBRAN,DIGS2
		DD	BRAN,DIGS1
DIGS2:		DD	EXIT

;   SIGN	( n -- )
;		Add a minus sign to the numeric output string.

		$COLON	4,'SIGN',SIGN
		DD	ZLESS
		DD	QBRAN,SIGN1
		DD	DOLIT,'-',HOLD
SIGN1:		DD	EXIT

;   #>		( w -- b u )
;		Prepare the output string to be TYPE'd.

		$COLON	2,'#>',EDIGS
		DD	DROP,HLD,AT
		DD	PAD,OVER,SUBB,EXIT

;   str		( w -- b u )
;		Convert a signed integer to a numeric string.

		$COLON	3,'str',STRI
		DD	DUPP,TOR,ABSS
		DD	BDIGS,DIGS,RFROM
		DD	SIGN,EDIGS,EXIT

;   HEX		( -- )
;		Use radix 16 as base for numeric conversions.

		$COLON	3,'HEX',HEX
		DD	DOLIT,16,BASE,STORE,EXIT

;   DECIMAL	( -- )
;		Use radix 10 as base for numeric conversions.

		$COLON	7,'DECIMAL',DECIM
		DD	DOLIT,10,BASE,STORE,EXIT

;; Numeric input, single precision

;   DIGIT?	( c base -- u t )
;		Convert a character to its numeric value. 
;		A true flag indicates success.

		$COLON	6,'DIGIT?',DIGTQ
		DD	TOR,DOLIT,'0',SUBB
		DD	DOLIT,9,OVER,LESS
		DD	QBRAN,DGTQ1
		DD	DOLIT,7,SUBB
		DD	DUPP,DOLIT,10,LESS,ORR
DGTQ1:		DD	DUPP,RFROM,ULESS,EXIT

;   NUMBER?	( a -- n T | a F )
;		Convert a number string to integer. Push a flag on tos.

		$COLON	7,'NUMBER?',NUMBQ
		DD	BASE,AT,TOR,DOLIT,0,OVER,COUNT
		DD	OVER,CAT,DOLIT,'$',EQUAL
		DD	QBRAN,NUMQ1
		DD	HEX,SWAP,ONEPLUS
		DD	SWAP,ONEMINUS
NUMQ1:		DD	OVER,CAT,DOLIT,'-',EQUAL,TOR
		DD	SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
		DD	QBRAN,NUMQ6
		DD	ONEMINUS,TOR
NUMQ2:		DD	DUPP,TOR,CAT,BASE,AT,DIGTQ
		DD	QBRAN,NUMQ4
		DD	SWAP,BASE,AT,STAR,PLUS,RFROM
		DD	ONEPLUS
		DD	DONXT,NUMQ2
		DD	RAT,SWAP,DROP
		DD	QBRAN,NUMQ3
		DD	NEGAT
NUMQ3:		DD	SWAP
		DD	BRAN,NUMQ5
NUMQ4:		DD	RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:		DD	DUPP
NUMQ6:		DD	RFROM,DDROP
		DD	RFROM,BASE,STORE,EXIT

;; Basic I/O

;   ?KEY	( -- c T | F )
;		Return input character and true, or a false if no input.

		$COLON	4,'?KEY',QKEY
		DD	TQKEY,ATEXE,EXIT

;   KEY		( -- c )
;		Wait for and return an input character.

		$COLON	3,'KEY',KEY
KEY1:		DD	QKEY
		DD	QBRAN,KEY1
		DD	EXIT

;   EMIT	( c -- )
;		Send a character to the output device.

		$COLON	4,'EMIT',EMIT
		DD	TEMIT,ATEXE,EXIT

;   KEYBOARD?	( -- t )
;		1 if terminal, 0 if redirected from file.

		$COLON	9,'KEYBOARD?',QKB
		DD	DOLIT,0,DOLIT,6,SYSCALL,EXIT

;   NUF?	( -- t )
;		Return false if no input, else pause and if CR return true.

		$COLON	4,'NUF?',NUFQ
		DD	QKB,DUPP,QBRAN,NUFQ1,DROP ; When using a terminal..
		DD	QKEY,DUPP
		DD	QBRAN,NUFQ1
		DD	DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:		DD	EXIT

;   PACE	( -- )
;		Send a pace character for the file downloading process.

		$COLON	4,'PACE',PACE
		DD	DOLIT,PACECHAR,EMIT,EXIT

;   SPACE	( -- )
;		Send the blank character to the output device.

		$COLON	5,'SPACE',SPACE
		DD	BLANK,EMIT,EXIT

;   SPACES	( +n -- )
;		Send n spaces to the output device.

		$COLON	6,'SPACES',SPACS
		DD	DOLIT,0,MAX,TOR
		DD	BRAN,CHAR2
CHAR1:		DD	SPACE
CHAR2:		DD	DONXT,CHAR1
		DD	EXIT

;   CR		( -- )
;		Output a carriage return and a line feed (DOS).
;		Output a line feed (UNIX).

		$COLON	2,'CR',CR
		DD	DOLIT,CRR,EMIT,EXIT

;   do$		( -- a )
;		Return the address of a compiled string.

		$COLON	COMPO+3,'do$',DOSTR
		DD	RFROM,RAT,RFROM,COUNT,PLUS
		DD	ALGND,TOR,SWAP,TOR,EXIT

;   $"|		( -- a )
;		Run time routine compiled by $". Return address of a compiled string.

		$COLON	COMPO+3,'$"|',STRQP
		DD	DOSTR,EXIT		; force a call to do$

;   ."|		( -- )
;		Run time routine of ." . Output a compiled string.

		$COLON	COMPO+3,'."|',DOTQP
		DD	DOSTR,COUNT,TYPES,EXIT

;   .R		( n +n -- )
;		Display an integer in a field of n columns, right justified.

		$COLON	2,'.R',DOTR
		DD	TOR,STRI,RFROM,OVER,SUBB
		DD	SPACS,TYPES,EXIT

;   U.R		( u +n -- )
;		Display an unsigned integer in n column, right justified.

		$COLON	3,'U.R',UDOTR
		DD	TOR,BDIGS,DIGS,EDIGS
		DD	RFROM,OVER,SUBB
		DD	SPACS,TYPES,EXIT

;   U.		( u -- )
;		Display an unsigned integer in free format.

		$COLON	2,'U.',UDOT
		DD	BDIGS,DIGS,EDIGS
		DD	SPACE,TYPES,EXIT

;   .		( w -- )
;		Display an integer in free format, preceeded by a space.

		$COLON	1,'.',DOT
		DD	BASE,AT,DOLIT,10,XORR	; ?decimal
		DD	QBRAN,DOT1
		DD	UDOT,EXIT		; no, display unsigned
DOT1:		DD	STRI,SPACE,TYPES,EXIT	; yes, display signed

;   ?		( a -- )
;		Display the contents in a memory cell.

		$COLON	1,'?',QUEST
		DD	AT,DOT,EXIT

;   TRUE	( -- -1 )
;		Required constant

		$COLON	IMEDD+4,'TRUE',TTRUE
		DD	DOLIT,-1
		DD	STATE,AT,QBRAN,ncompt
		DD	DOLIT,DOLIT,COMPIC,COMMA
ncompt:		DD	EXIT

;   FALSE	( -- 0 )
;		Required constant

		$COLON	IMEDD+5,'FALSE',FFALSE
		DD	DOLIT,0
		DD	STATE,AT,QBRAN,ncompf
		DD	DOLIT,DOLIT,COMPIC,COMMA
ncompf:		DD	EXIT


;; Parsing

;   parse	( b u c -- b u delta ; <string> )
;		Scan string delimited by c. Return found string and its offset.

		$COLON	5,'parse',PARS
		DD	TEMP,STORE,OVER,TOR,DUPP
		DD	QBRAN,PARS8
		DD	ONEMINUS,TEMP,AT,BLANK,EQUAL
		DD	QBRAN,PARS3
		DD	TOR
PARS1:		DD	BLANK,OVER,CAT		; skip leading blanks ONLY
		DD	SUBB,ZLESS,INVER
		DD	QBRAN,PARS2
		DD	ONEPLUS
		DD	DONXT,PARS1
		DD	RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:		DD	RFROM
PARS3:		DD	OVER,SWAP
		DD	TOR
PARS4:		DD	TEMP,AT,OVER,CAT,SUBB	; scan for delimiter
		DD	TEMP,AT,BLANK,EQUAL
		DD	QBRAN,PARS5
		DD	ZLESS
PARS5:		DD	QBRAN,PARS6
		DD	ONEPLUS
		DD	DONXT,PARS4
		DD	DUPP,TOR
		DD	BRAN,PARS7
PARS6:		DD	RFROM,DROP,DUPP
		DD	ONEPLUS,TOR
PARS7:		DD	OVER,SUBB
		DD	RFROM,RFROM,SUBB,EXIT
PARS8:		DD	OVER,RFROM,SUBB,EXIT

;   PARSE	( c -- b u ; <string> )
;		Scan input stream and return counted string delimited by c.

		$COLON	5,'PARSE',PARSE
		DD	TOR,TIB,INN,AT,PLUS	; current input buffer pointer
		DD	NTIB,AT,INN,AT,SUBB	; remaining count
		DD	RFROM,PARS,INN,PSTOR,EXIT

;   .(		( -- )
;		Output following string up to next ) .

		$COLON	IMEDD+2,'.(',DOTPR
		DD	DOLIT,')',PARSE,TYPES,EXIT

;   (		( -- )
;		Ignore following string up to next ) . A comment.

		$COLON	IMEDD+1,'(',PAREN
		DD	DOLIT,')',PARSE,DDROP,EXIT

;   \		( -- )
;		Ignore following text till the end of line.

		$COLON	IMEDD+1,'\',BKSLA
		DD	NTIB,AT,INN,STORE,EXIT

;   CHAR	( -- c )
;		Parse next word and return its first character.

		$COLON	4,'CHAR',CHAR
		DD	BLANK,PARSE,DROP,CAT,EXIT

;   [CHAR]	( -- ch | -- )
;		Compile the first char of the following string

		$COLON	IMEDD+6,'[CHAR]',BCHARB
		DD	CHAR,LITER,EXIT

;   TOKEN	( -- a ; <string> )
;		Parse a word from input stream and copy it to name dictionary.
;		The string at a STOPS at an aligned address.

		$COLON	5,'TOKEN',TOKEN
		DD	BLANK,PARSE,DOLIT,31,MIN
		DD	NP,AT,OVER,SUBB,CELLM
		DD	PACKS,EXIT

;   WORD	( c -- a ; <string> )
;		Parse a word from input stream and copy it to code dictionary.

		$COLON	4,'WORD',WORDD
		DD	PARSE,HERE,PACKS,EXIT

;; Dictionary search

;   NAME>	( na -- ca )
;		Return a code address given a name address.

		$COLON	5,'NAME>',NAMET
		DD	CELLM,CELLM,AT,EXIT

;   SAME?	( a a u -- a a f \ -0+ )
;		Compare u cells in two strings. Return 0 if identical.

		$COLON	5,'SAME?',SAMEQ
		DD	TOR
		DD	BRAN,SAME2
SAME1:		DD	OVER,RAT,CELLS,PLUS,AT
		DD	OVER,RAT,CELLS,PLUS,AT
		DD	SUBB,QDUP
		DD	QBRAN,SAME2
		DD	RFROM,DROP,EXIT
SAME2:		DD	DONXT,SAME1
		DD	DOLIT,0,EXIT

;   find	( a va -- ca na | a F )
;		Search a vocabulary for a string. 
;		Return ca and na if succeeded.

		$COLON	4,'find',FIND
		DD	SWAP,DUPP,CAT
		DD	DOLIT,CELLL,SLASH,TEMP,STORE
		DD	DUPP,AT,TOR,CELLP,SWAP
FIND1:		DD	AT,DUPP
		DD	QBRAN,FIND6
		DD	DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
		DD	QBRAN,FIND2
		DD	CELLP,DOLIT,-1
		DD	BRAN,FIND3
FIND2:		DD	CELLP,TEMP,AT,SAMEQ
FIND3:		DD	BRAN,FIND4
FIND6:		DD	RFROM,DROP
		DD	SWAP,CELLM,SWAP,EXIT
FIND4:		DD	QBRAN,FIND5
		DD	CELLM,CELLM
		DD	BRAN,FIND1
FIND5:		DD	RFROM,DROP,SWAP,DROP
		DD	CELLM
		DD	DUPP,NAMET,SWAP,EXIT

;   NAME?	( a -- ca na | a F )
;		Search all context vocabularies for a string.

		$COLON	5,'NAME?',NAMEQ
		DD	CNTXT,DUPP,DAT,XORR
		DD	QBRAN,NAMQ1
		DD	CELLM
NAMQ1:		DD	TOR
NAMQ2:		DD	RFROM,CELLP,DUPP,TOR
		DD	AT,QDUP
		DD	QBRAN,NAMQ3
		DD	FIND,QDUP
		DD	QBRAN,NAMQ2
		DD	RFROM,DROP,EXIT
NAMQ3:		DD	RFROM,DROP
		DD	DOLIT,0,EXIT

;; Terminal response

;   ^H		( bot eot cur -- bot eot cur )
;		Backup the cursor by one character.

		$COLON	2,'^H',BKSP
		DD	TOR,OVER,RFROM,SWAP,OVER,XORR
		DD	QBRAN,BACK1
		DD	DOLIT,BKSPP,TECHO,ATEXE,ONEMINUS
		DD	BLANK,TECHO,ATEXE
		DD	DOLIT,BKSPP,TECHO,ATEXE
BACK1:		DD	EXIT

;   TAP		( bot eot cur c -- bot eot cur )
;		Accept and echo the key stroke and bump the cursor.

		$COLON	3,'TAP',TAP
		DD	DUPP,TECHO,ATEXE
		DD	OVER,CSTOR,ONEPLUS,EXIT

;   kTAP	( bot eot cur c -- bot eot cur )
;		Process a key stroke, CR or backspace.

		$COLON	4,'kTAP',KTAP
		DD	DUPP,DOLIT,CRR,XORR
		DD	QBRAN,KTAP2
		DD	DOLIT,BKSPP,XORR
		DD	QBRAN,KTAP1
		DD	BLANK,TAP,EXIT
KTAP1:		DD	BKSP,EXIT
KTAP2:		DD	DROP,SWAP,DROP,DUPP,EXIT

;   accept	( b u -- b u )
;		Accept characters to input buffer. Return with actual count.

		$COLON	6,'accept',ACCEP
		DD	OVER,PLUS,OVER
ACCP1:		DD	DDUP,XORR
		DD	QBRAN,ACCP4
		DD	KEY,DUPP
		DD	BLANK,DOLIT,127,WITHI
		DD	QBRAN,ACCP2
		DD	TAP
		DD	BRAN,ACCP3
ACCP2:		DD	TTAP,ATEXE
ACCP3:		DD	BRAN,ACCP1
ACCP4:		DD	DROP,OVER,SUBB,EXIT

;   EXPECT	( b u -- )
;		Accept input stream and store count in SPAN.

		$COLON	6,'EXPECT',EXPEC
		DD	TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;   QUERY	( -- )
;		Accept input stream to terminal input buffer.

		$COLON	5,'QUERY',QUERY
		DD	TIB,DOLIT,RTS,TEXPE,ATEXE,NTIB,STORE
		DD	DROP,DOLIT,0,INN,STORE,EXIT

;; Error handling

;   CATCH	( ca -- 0 | err# )
;		Execute word at ca and set up an error frame for it.

		$COLON	5,'CATCH',CATCH
		DD	SPAT,TOR,HANDL,AT,TOR	; save error frame
		DD	RPAT,HANDL,STORE,EXECU	; execute
		DD	RFROM,HANDL,STORE	; restore error frame
		DD	RFROM,DROP,DOLIT,0,EXIT	; no error

;   THROW	( err# -- err# )
;		Reset system to current local error frame an update error flag.

		$COLON	5,'THROW',THROW
		DD	HANDL,AT,RPSTO		; restore return stack
		DD	RFROM,HANDL,STORE	; restore handler frame
		DD	RFROM,SWAP,TOR,SPSTO	; restore data stack
		DD	DROP,RFROM,EXIT

;   NULL$	( -- a )
;		Return address of a null string with zero count.

		$COLON	5,'NULL$',NULLS
		DD	DOVAR			; emulate CREATE
		DD	0
		DB	'coyote',0,0

;   ABORT	( -- )
;		Reset data stack and jump to QUIT.

		$COLON	5,'ABORT',ABORT
		DD	NULLS,THROW

;   abort"	( f -- )
;		Run time routine of ABORT" . Abort with a message.

		$COLON	COMPO+6,'abort"',ABORQ
		DD	QBRAN,ABOR1		; text flag
		DD	DOSTR,THROW		; pass error string
ABOR1:		DD	DOSTR,DROP,EXIT		; drop error

;; The text interpreter

;   $INTERPRET	( a -- )
;		Interpret a word. If failed, try to convert it to an integer.

		$COLON	10,'$INTERPRET',INTER
		DD	NAMEQ,QDUP		; ?defined
		DD	QBRAN,INTE1
		DD	AT,DOLIT,COMPO,ANDD	; ?compile only lexicon bits
		DD	ABORQ
		DB	13,' compile only',0,0
		DD	EXECU,EXIT		; execute defined word
INTE1:		DD	TNUMB,ATEXE		; convert a number
		DD	QBRAN,INTE2
		DD	EXIT
INTE2:		DD	THROW			; error

;   [		( -- )
;		Start the text interpreter.

		$COLON	IMEDD+1,'[',LBRAC
		DD	DOLIT,INTER,TEVAL,STORE
		DD	DOLIT,FALSEE,STATE,STORE,EXIT

;   .OK		( -- )
;		Display 'ok' only while interpreting.

		$COLON	3,'.OK',DOTOK
		DD	DOLIT,INTER,TEVAL,AT,EQUAL
		DD	QBRAN,DOTO1
		DD	DOTQP
		DB	3,' ok'
DOTO1:		DD	CR,EXIT

;   ?STACK	( -- )
;		Abort if the data stack underflows.

		$COLON	6,'?STACK',QSTAC
		DD	DEPTH,ZLESS		; check only for underflow
		DD	ABORQ
		DB	10,' underflow',0
		DD	EXIT

;   EVAL	( -- )
;		Interpret the input stream.

		$COLON	4,'EVAL',EVAL
EVAL1:		DD	TOKEN,DUPP,CAT		; ?input stream empty
		DD	QBRAN,EVAL2
		DD	TEVAL,ATEXE,QSTAC	; evaluate input, check stack
		DD	BRAN,EVAL1
EVAL2:		DD	DROP,TPROM,ATEXE,EXIT	; prompt

;; Shell

;   PRESET	( -- )
;		Reset data stack pointer and the terminal input buffer.

		$COLON	6,'PRESET',PRESE
		DD	SZERO,AT,SPSTO
		DD	DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;   xio		( a a a -- )
;		Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

		$COLON	COMPO+3,'xio',XIO
		DD	DOLIT,ACCEP,TEXPE,DSTOR
		DD	TECHO,DSTOR,EXIT

;   FILE	( -- )
;		Select I/O vectors for file download.

		$COLON	4,'FILE',FILE
		DD	DOLIT,PACE,DOLIT,DROP
		DD	DOLIT,KTAP,XIO,EXIT

;   HAND	( -- )
;		Select I/O vectors for terminal interface.

		$COLON	4,'HAND',HAND
		DD	DOLIT,DOTOK,DOLIT,EMIT
		DD	DOLIT,KTAP,XIO,EXIT

;   I/O		( -- a )
;		Array to store default I/O vectors.

		$COLON	3,'I/O',ISLO
		DD	DOVAR			; emulate CREATE
		DD	QRX,TXSTO		; default I/O vectors

;   CONSOLE	( -- )
;		Initiate terminal interface.

		$COLON	7,'CONSOLE',CONSO
		DD	ISLO,DAT,TQKEY,DSTOR	; restore default I/O device
		DD	HAND,EXIT		; keyboard input

;   QUIT	( -- )
;		Reset return stack pointer and start text interpreter.

		$COLON	4,'QUIT',QUIT
		DD	RZERO,AT,RPSTO		; reset return stack pointer
QUIT1:		DD	LBRAC			; start interpretation
QUIT2:		DD	QUERY			; get input
		DD	DOLIT,EVAL,CATCH,QDUP	; evaluate input
		DD	QBRAN,QUIT2		; continue till error
		DD	TPROM,AT,TOR		; save input device
		DD	CONSO,NULLS,OVER,XORR	; ?display error message
		DD	QBRAN,QUIT3
		DD	SPACE,COUNT,TYPES	; error message
		DD	DOTQP
		DB	3,' ? '			; error prompt
QUIT3:		DD	RFROM,DOLIT,DOTOK,XORR	; ?file input
		DD	QBRAN,QUIT4
		DD	DOLIT,ERRO,EMIT		; file error, tell host
QUIT4:		DD	PRESE			; some cleanup
		DD	BRAN,QUIT1

;; The compiler

;   '		( -- ca )
;		Search context vocabularies for the next word in input stream.

		$COLON	1,"'",TICK
		DD	TOKEN,NAMEQ		; ?defined
		DD	QBRAN,TICK1
		DD	EXIT			; yes, push code address
TICK1:		DD	THROW			; no, error

;   ALLOT	( n -- )
;		Allocate n bytes to the code dictionary.

		$COLON	5,'ALLOT',ALLOT
		DD	CP,PSTOR,EXIT		; adjust code pointer

;   ,		( w -- )
;		Compile an integer into the code dictionary.

		$COLON	1,',',COMMA
		DD	HERE,DUPP,CELLP		; cell boundary
		DD	CP,STORE,STORE,EXIT	; adjust code pointer and compile

;   COMPILE,	( w -- )
;		Compile an execution token into the code dictionary.

		$COLON	COMPO+8,'COMPILE,',COMPIC
		DD	lastt,AT,penult,STORE
		DD	DUPP,lastt,STORE
		DD	HERE,DUPP,CELLP		; cell boundary
		DD	CP,STORE,STORE,EXIT	; adjust code pointer and compile

;   [COMPILE]	( -- ; <string> )
;		Compile the next immediate word into code dictionary.

		$COLON	IMEDD+9,'[COMPILE]',BCOMP
		DD	TICK,COMPIC,EXIT

;   COMPILE	( -- )
;		Compile the next address in colon list to code dictionary.

		$COLON	COMPO+7,'COMPILE',COMPI
		DD	RFROM,DUPP,AT,COMPIC	; compile address
		DD	CELLP,TOR,EXIT		; adjust return address

;   LITERAL	( w -- )
;		Compile tos to code dictionary as an integer literal.

		$COLON	IMEDD+7,'LITERAL',LITER
		DD	COMPI,DOLIT,COMMA,EXIT

;   $,"		( -- )
;		Compile a literal string up to next " .

		$COLON	3,'$,"',STRCQ
		DD	DOLIT,'"',WORDD		; move string to code dictionary
		DD	COUNT,PLUS,ALGND	; calculate aligned end of string
		DD	CP,STORE,EXIT		; adjust the code pointer

;   S"		( -- c-addr u | -- )
;		compiling: Compile a literal string up to next " .
;		executing: give address and count of following string

		$COLON	IMEDD+2,'S"',SQUOTE
		DD	DOLIT,INTER,TEVAL,AT,EQUAL
		DD	QBRAN,comping
		DD	DOLIT,'"',WORDD,COUNT,EXIT	; exec
comping:	DD	STRQ,DOLIT,COUNT,COMMA,EXIT	; compile

;   RECURSE	( -- )
;		Make the current word available for compilation.

		$COLON	IMEDD+7,'RECURSE',RECUR
		DD	LAST,AT,NAMET,COMPIC,EXIT

;; Structures

;   ?branch,	( -- )
; 		Compile conditional branch code. Vectored through (?BRANCH).

		$COLON	8,'?branch,',qbrancom
		DD	bqbran,ATEXE,EXIT

;   FOR		( -- a )
;		Start a FOR-NEXT loop structure in a colon definition.

		$COLON	IMEDD+3,'FOR',FOR
		DD	COMPI,TOR,HERE,EXIT

;   BEGIN	( -- a )
;		Start an infinite or indefinite loop structure.

		$COLON	IMEDD+5,'BEGIN',BEGIN
		DD	HERE,EXIT

;   NEXT	( a -- )
;		Terminate a FOR-NEXT loop structure.

		$COLON	IMEDD+4,'NEXT',NEXT
		DD	COMPI,DONXT,COMMA,EXIT

;   UNTIL	( a -- )
;		Terminate a BEGIN-UNTIL indefinite loop structure.

		$COLON	IMEDD+5,'UNTIL',UNTIL
		DD	qbrancom,COMMA,EXIT

;   AGAIN	( a -- )
;		Terminate a BEGIN-AGAIN infinite loop structure.

		$COLON	IMEDD+5,'AGAIN',AGAIN
		DD	COMPI,BRAN,COMMA,EXIT

;   IF		( -- A )
;		Begin a conditional branch structure.

		$COLON	IMEDD+2,'IF',IFF
		DD	qbrancom,HERE
		DD	DOLIT,0,COMMA,EXIT

;   ORIF	( -- A )
;		Begin a conditional branch structure.

		$COLON	IMEDD+4,'ORIF',ORIFF
		DD	COMPI,QDUP,COMPI,INVER
		DD	qbrancom,HERE
		DD	DOLIT,0,COMMA,EXIT

;   ANDIF	( -- A )
;		Begin a conditional branch structure.

		$COLON	IMEDD+5,'ANDIF',ANDIFF
		DD	COMPI,DUPP
		DD	qbrancom,HERE
		DD	DOLIT,0,COMMA
		DD	COMPI,DROP,EXIT

;   AHEAD	( -- A )
;		Compile a forward branch instruction.

		$COLON	IMEDD+5,'AHEAD',AHEAD
		DD	COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT	( A a -- )
;		Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		$COLON	IMEDD+6,'REPEAT',REPEA
		DD	AGAIN,HERE,SWAP,STORE,EXIT

;   THEN	( A -- )
;		Terminate a conditional branch structure.

		$COLON	IMEDD+4,'THEN',THENN
		DD	HERE,SWAP,STORE,EXIT

;   AFT		( a -- a A )
;		Jump to THEN in a FOR-AFT-THEN-NEXT loop 
;		the first time through.

		$COLON	IMEDD+3,'AFT',AFT
		DD	DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE	( A -- A )
;		Start the false clause in an IF-ELSE-THEN structure.

		$COLON	IMEDD+4,'ELSE',ELSEE
		DD	AHEAD,SWAP,THENN,EXIT

;   WHILE	( a -- A a )
;		Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		$COLON	IMEDD+5,'WHILE',WHILEE
		DD	IFF,SWAP,EXIT

;   DO		( -- a )
;		Start a DO-LOOP loop structure in a colon definition.

		$COLON	IMEDD+2,'DO',DODO
		DD	COMPI,SWAP,COMPI,TOR,COMPI,TOR,HERE,EXIT

;   LOOP	( a -- )
;		Terminate a DO LOOP loop structure.

		$COLON	IMEDD+4,'LOOP',LOOPP
		DD	COMPI,DOLP,COMMA,EXIT

;   +LOOP	( a -- )
;		Terminate a DO +LOOP loop structure.

		$COLON	IMEDD+5,'+LOOP',PLOOPP
		DD	COMPI,DOPLP,COMMA,EXIT

;   DUP+LOOP	( a -- )
;		Terminate a DO +LOOP loop structure.

		$COLON	IMEDD+8,'DUP+LOOP',DPLOOPP
		DD	COMPI,DODPLP,COMMA,EXIT

;   ABORT"	( -- ; <string> )
;		Conditional abort with an error message.

		$COLON	IMEDD+6,'ABORT"',ABRTQ
		DD	COMPI,ABORQ,STRCQ,EXIT

;   $"		( -- ; <string> )
;		Compile an inline string literal.

		$COLON	IMEDD+2,'$"',STRQ
		DD	COMPI,STRQP,STRCQ,EXIT

;   ."		( -- ; <string> )
;		Compile an inline string literal to be typed out at run time.

		$COLON	IMEDD+2,'."',DOTQ
		DD	COMPI,DOTQP,STRCQ,EXIT

;; Name compiler

;   ?UNIQUE	( a -- a )
;		Display a warning message if the word already exists.

		$COLON	7,'?UNIQUE',UNIQU
		DD	DUPP,NAMEQ		; ?name exists
		DD	QBRAN,UNIQ1
		DD	DOTQP			; redefinitions are OK
		DB	7,' reDef '		; but the user should be warned
		DD	OVER,COUNT,TYPES	; just in case it's not planned
UNIQ1:		DD	DROP,EXIT

;   $,n		( na -- )
;		Build a new dictionary name using the string at na.

		$COLON	3,'$,n',SNAME
		DD	DUPP,CAT		; ?null input
		DD	QBRAN,PNAM1
		DD	UNIQU			; ?redefinition
		DD	DUPP,LAST,STORE		; save na for vocabulary link
;		DD	HERE,ALGND,SWAP		; align code address
		DD	ALGN,HERE,SWAP		; make SURE it is aligned
		DD	CELLM			; link address
		DD	CRRNT,AT,AT,OVER,STORE
		DD	CELLM,DUPP,NP,STORE	; adjust name pointer
		DD	STORE,EXIT		; save code pointer
PNAM1:		DD	STRQP
		DB	5,' name',0,0		; null input
		DD	THROW

;; FORTH compiler

;   $COMPILE	( a -- )
;		Compile next word to code dictionary as a token or literal.

		$COLON	8,'$COMPILE',SCOMP
		DD	NAMEQ,QDUP		; ?defined
		DD	QBRAN,SCOM2
		DD	AT,DOLIT,IMEDD,ANDD	; ?immediate
		DD	QBRAN,SCOM1
		DD	EXECU,EXIT		; it's immediate, execute
SCOM1:		DD	COMPIC,EXIT		; it's not immediate, compile
SCOM2:		DD	TNUMB,ATEXE		; try to convert to number
		DD	QBRAN,SCOM3
		DD	LITER,EXIT		; compile number as integer
SCOM3:		DD	THROW			; error

;   OVERT	( -- )
;		Link a new word into the current vocabulary.

		$COLON	5,'OVERT',OVERT
		DD	LAST,AT,CRRNT,AT,STORE,EXIT

;   ;		( -- )
;		Terminate a colon definition.

		$COLON	IMEDD+COMPO+1,';',SEMIS
		DD	COMPI,EXIT,COMPI,EXITT,LBRAC,OVERT,EXIT

;   ]		( -- )
;		Start compiling the words in the input stream.

		$COLON	1,']',RBRAC
		DD	DOLIT,SCOMP,TEVAL,STORE
		DD	DOLIT,TRUEE,STATE,STORE,EXIT

;   call,	( ca -- )
;		Assemble a call instruction to ca.

		$COLON	5,'call,',CALLC
		DD	DOLIT,CALLL,COMMA,HERE	;Direct Threaded Code
		DD	CELLP,SUBB,COMMA,EXIT	;DTC 80386 relative call

;   :		( -- ; <string> )
;		Start a new colon definition using next word as its name.

		$COLON	1,':',COLON
		DD	TOKEN,SNAME,DOLIT,DOLST
		DD	CALLC,RBRAC,EXIT

;   IMMEDIATE	( -- )
;		Make the last compiled word an immediate word.

		$COLON	9,'IMMEDIATE',IMMED
		DD	DOLIT,IMEDD,LAST,AT,AT,ORR
		DD	LAST,AT,STORE,EXIT

;; Defining words

;   USER	( u -- ; <string> )
;		Compile a new user variable.

		$COLON	4,'USER',USER
		DD	TOKEN,SNAME,OVERT
		DD	DOLIT,DOLST,CALLC
		DD	DOLIT,DOUSE,COMMA
		DD	COMMA,EXIT

;   CONSTANT	( u -- ; <string> )
;		Compile a constant

		$COLON	8,'CONSTANT',CONS
		DD	TOKEN,SNAME,OVERT,IMMED
		DD	DOLIT,DOLST,CALLC
		DD	DOLIT,SDOCONS,COMMA
		DD	COMMA,EXIT

;   VARIABLE	( -- ; <string> )
;		Compile a new variable initialized to 0.

		$COLON	8,'VARIABLE',VARIA
		DD	TOKEN,SNAME,OVERT,IMMED
		DD	DOLIT,DOLST,CALLC
		DD	DOLIT,SDOVAR,COMMA
		DD	DOLIT,0,COMMA,EXIT


;   CREATE	( -- ; <string> )
;		Compile a new array entry without allocating code space.

		$COLON	6,'CREATE',CREAT
		DD	TOKEN,SNAME,OVERT
		DD	DOLIT,DOLST,CALLC
		DD	DOLIT,DOVAR,COMMA,EXIT

;  dodoes	( -- )
		$COLON	COMPO+6,'dodoes',ddoes
		DD	CP,AT
		DD	LAST,AT,NAMET,DOLIT,2*CELLL,PLUS,CP,STORE
		DD	RFROM,COMMA
		DD	CP,STORE,EXIT

;   DOES>	( -- addr )
;		modify latest created word to jump into the DOES part of a
;		defining word with its pfa on the data stack

		$COLON	IMEDD+5,'DOES>',DOES
		DD	DOLIT,ddoes,COMMA
		DD	DOLIT,DODOIT,CALLC,EXIT

;; Tools

;   _TYPE	( b u -- )
;		Display a string. Filter non-printing characters.

		$COLON	5,'_TYPE',UTYPE
		DD	TOR			; start count down loop
		DD	BRAN,UTYP2		; skip first pass
UTYP1:		DD	DUPP,CAT,TCHAR,EMIT	; display only printable
		DD	ONEPLUS			; increment address
UTYP2:		DD	DONXT,UTYP1		; loop till done
		DD	DROP,EXIT

;   dm+		( a u -- a )
;		Dump u bytes from , leaving a+u on the stack.

		$COLON	3,'dm+',DUMPP
		DD	OVER,DOLIT,4,UDOTR	; display address
		DD	SPACE,TOR		; start count down loop
		DD	BRAN,PDUM2		; skip first pass
PDUM1:		DD	DUPP,CAT,DOLIT,3,UDOTR	; display numeric data
		DD	ONEPLUS			; increment address
PDUM2:		DD	DONXT,PDUM1		; loop till done
		DD	EXIT

;   DUMP	( a u -- )
;		Dump u bytes from a, in a formatted manner.

		$COLON	4,'DUMP',DUMP
		DD	BASE,AT,TOR,HEX		; save radix, set hex
		DD	DOLIT,16,SLASH		; change count to lines
		DD	TOR			; start count down loop
DUMP1:		DD	CR,DOLIT,16,DDUP,DUMPP	; display numeric
		DD	ROT,ROT
		DD	DOLIT,2,SPACS,UTYPE	; display printable characters
		DD	NUFQ,INVER		; user control
		DD	QBRAN,DUMP2
		DD	DONXT,DUMP1		; loop till done
		DD	BRAN,DUMP3
DUMP2:		DD	RFROM,DROP		; cleanup loop stack, early exit
DUMP3:		DD	DROP,RFROM,BASE,STORE	; restore radix
		DD	EXIT

;   .S		( ... -- ... )
;		Display the contents of the data stack.

		$COLON	2,'.S',DOTS
		DD	CR,DEPTH		; stack depth
		DD	TOR			; start count down loop
		DD	BRAN,DOTS2		; skip first pass
DOTS1:		DD	RAT,PICK,DOT		; index stack, display contents
DOTS2:		DD	DONXT,DOTS1		; loop till done
		DD	DOTQP
		DB	4,' <sp',0,0,0
		DD	EXIT

;   !CSP	( -- )
;		Save stack pointer in CSP for error checking.

		$COLON	4,'!CSP',STCSP
		DD	SPAT,CSP,STORE,EXIT	; save pointer

;   ?CSP	( -- )
;		Abort if stack pointer differs from that saved in CSP.

		$COLON	4,'?CSP',QCSP
		DD	SPAT,CSP,AT,XORR	; compare pointers
		DD	ABORQ			; abort if different
		DB	6,'stacks',0
		DD	EXIT

;   >NAME	( ca -- na | F )
;		Convert code address to a name address.

		$COLON	5,'>NAME',TNAME
	DD	CRRNT			; current		; vocabulary link
TNAM1:	DD	CELLP,AT,QDUP		;begin cell+ @ ?dup	; check all vocabularies
	DD	QBRAN,TNAM4		;while
	DD	DDUP			; 2dup
TNAM2:	DD	AT,DUPP			; begin @ dup		; ?last word in a vocabulary
	DD	QBRAN,TNAM3		; while
	DD	DDUP,NAMET,XORR		;  2dup name> xor	; compare
	DD	QBRAN,TNAM3		; while
	DD	CELLM			;  cell-		; continue with next word
	DD	BRAN,TNAM2		; repeat
TNAM3:	DD	SWAP,DROP,QDUP		; then swap drop ?dup
	DD	QBRAN,TNAM1		;until
	DD	SWAP,DROP,SWAP,DROP,EXIT; -rot 2drop ;s
TNAM4:	DD	DROP,DOLIT,0,EXIT	;then drop 0 ;s

;   .ID		( na -- )
;		Display the name at address.

		$COLON	3,'.ID',DOTID
		DD	QDUP			; if zero no name
		DD	QBRAN,DOTI1
		DD	COUNT,DOLIT,01FH,ANDD	; mask lexicon bits
		DD	UTYPE,EXIT		; display name string
DOTI1:		DD	DOTQP
		DB	9,' {noName}',0,0
		DD	EXIT

;   SEE		( -- ; <string> )
;		A simple decompiler.

		$COLON	3,'SEE',SEE
		DD	TICK			; starting address
		DD	CR,CELLP
SEE1:		DD	CELLP,DUPP,AT,DUPP	; ?does it contain a zero
		DD	QBRAN,SEE2
		DD	TNAME			; ?is it a name
SEE2:		DD	QDUP			; name address or zero
		DD	QBRAN,SEE3
		DD	SPACE,DOTID		; display name
		DD	BRAN,SEE4
SEE3:		DD	DUPP,AT			; display number
		DD	DUPP,DOLIT,EXITT,NEQUAL,QBRAN,SEE5
		DD	UDOT
		DD	BRAN,SEE4
SEE5:		DD	DROP
SEE4:		DD	NUFQ			; user control
		DD	QBRAN,SEE1
		DD	DROP,EXIT

;   WORDS	( -- )
;		Display the names in the context vocabulary.

		$COLON	5,'WORDS',WORDS
		DD	CR,CNTXT,AT		; only in context
WORS1:		DD	AT,QDUP			; ?at end of list
		DD	QBRAN,WORS2
		DD	DUPP,SPACE,DOTID	; display a name
		DD	CELLM,NUFQ		; user control
		DD	QBRAN,WORS1
		DD	DROP
WORS2:		DD	EXIT

;; Hardware reset

;   VER		( -- n )
;		Return the version number of this implementation.

		$COLON	3,'VER',VERSN
		DD	DOLIT,VER*256+EXT,EXIT

;   hi		( -- )
;		Display the sign-on message of eForth.

		$COLON	2,'hi',HI
		DD	STOIO,CR,DOTQP		; initialize I/O
		DB	11,'eForth v'		; model
		DB	VER+'0','.',EXT+'0'	; version
		DD	CR,EXIT

;   'BOOT	( -- a )
;		The application startup vector.

		$COLON	5,"'BOOT",TBOOT
		DD	DOVAR
		DD	HI			; application to boot

;   COLD	( -- )
;		The hilevel cold start sequence.

		$COLON	4,'COLD',COLD
COLD1:		DD	DOLIT,UZERO,DOLIT,UPP
		DD	DOLIT,ULAST-UZERO,CMOVE	; initialize user area
		DD	PRESE			; initialize data stack and TIB
		DD	TBOOT,ATEXE		; application boot
		DD	FORTH,CNTXT,AT,DUPP	; initialize search order
		DD	CRRNT,DSTOR,OVERT
		DD	QUIT			; start interpretation
		DD	BRAN,COLD1		; just in case

;=============================================================================
;; Memory allocation

CTOP		EQU $			; next available code dictionary loc.
LASTN		EQU _NAME+2*CELLL+ORIG	; last name address in name dictionary
NTOP		EQU _NAME+ORIG		; next available name dictionary loc.

		ORG	DICTSIZE+ORIG

SPP:		DB 8*CELLL  DUP (0)	; start of data stack (SP0)
TIBB:		DB   RTS    DUP (0)	; terminal input buffer (TIB)
		DB   RTS    DUP (0)	; return stack (RP0)
RPP:		DB 8*CELLL  DUP (0)	; start of return stack (RP0)
UPP:		DB   US     DUP (0)	; user area (UP0)
 		DB NAMESIZE DUP (?)	; name dictionary

EFSEG		ENDS
END 		ORIG
;==============================================================================

